@@ -103,13 +103,17 @@ def sync(config, state, catalog):
     all_sub_stream_ids = get_sub_stream_ids()
 
     # Loop over streams in catalog
-    for stream in catalog.streams:
+    for stream in catalog.get_selected_streams(state):
         stream_id = stream.tap_stream_id
         mdata = metadata.to_map(stream.metadata)
         if stream_id not in selected_stream_ids:
             LOGGER.info("%s: Skipping - not selected", stream_id)
             continue            # TODO: sync code for stream goes here...
 
+        # Parent stream will sync sub stream
+        if stream_id in all_sub_stream_ids:
+            continue
+
         LOGGER.info('START Syncing: %s', stream_id)
         update_currently_syncing(state, stream_id)
 
@@ -118,7 +122,7 @@ def sync(config, state, catalog):
         singer.write_schema(
             stream_id, stream.schema.to_dict(), key_properties)
 
-        sub_stream_ids = SUB_STREAMS.get(stream_id)
+        sub_stream_ids = [SUB_STREAMS.get(stream_id)]
 
         # Populate class schemas and write a schema for the selected substreams of the stream
         if sub_stream_ids:
@@ -133,9 +137,6 @@ def sync(config, state, catalog):
                 singer.write_schema(
                     sub_stream.tap_stream_id, sub_stream.schema.to_dict(), sub_key_properties)
 
-        # parent stream will sync sub stream
-        if stream_id in all_sub_stream_ids:
-            continue
 
         LOGGER.info("Stream %s: Starting sync", stream_id)
         instance = STREAMS[stream_id](config) # Intialize class for selected stream
