@@ -1,8 +1,9 @@
-# pylint: disable=E1101,R0201,W0613
+# pylint: disable=E1101,R0201,R0904,W0222,W0613
 
 #!/usr/bin/env python3
 import json
 import os
+import sys
 import time
 from datetime import datetime, timedelta, timezone
 from ast import literal_eval
@@ -32,7 +33,7 @@ session = requests.Session()
 REQUEST_TIMEOUT = 300
 DEFAULT_INCLUDE_ANONYMOUS_VISITORS = 'false'
 
-BACKOFF_FACTOR = 30
+BACKOFF_FACTOR = 5
 
 def to_giveup(error):
     """
@@ -88,6 +89,18 @@ def update_currently_syncing(state, stream_name):
     singer.write_state(state)
 
 
+def retry_handler(details):
+    """Sets request retry count which is used to set request timeout in subsequent request"""
+    _, exception, _ = sys.exc_info()
+    if isinstance(exception, (ReadTimeoutError, requests.exceptions.RequestException)):
+        Stream.request_retry_count = details['tries'] + 1
+
+
+def reset_request_retry_count(details):
+    """Reset the Stream retry count in case we continue execution even after max. retries"""
+    Stream.request_retry_count = 1
+
+
 class Server42xRateLimitError(Exception):
     pass
 
@@ -126,19 +139,38 @@ class Stream():
     key_properties = KEY_PROPERTIES
     stream = None
     period = None
+    request_retry_count = 1
+    last_processed = None
+
     # initialized the endpoint attribute which can be overriden by child streams based on
     # the different parameters used by the stream.
     endpoint = Endpoints("/api/v1/aggregation", "POST")
 
     def __init__(self, config=None):
         self.config = config
+        self.record_limit = self.get_default_record_limit()
 
-    def send_request_get_results(self, req, endpoint, params, count, **kwargs):
-        # Set request timeout to config param `request_timeout` value.
         # If value is 0,"0", "" or None then it will set default to default to 300.0 seconds if not passed in config.
         config_request_timeout = self.config.get('request_timeout')
-        request_timeout = config_request_timeout and float(config_request_timeout) or REQUEST_TIMEOUT # pylint: disable=consider-using-ternary
-        resp = session.send(req, timeout=request_timeout)
+        if config_request_timeout and float(config_request_timeout) > 0:
+            self.request_timeout = float(config_request_timeout)
+        else:
+            self.request_timeout = REQUEST_TIMEOUT
+
+    def get_default_record_limit(self):
+        # Record limit will throttle the number of records getting replicated
+        # This limit will resolve request timeouts and will reduce the peak memory consumption
+        record_limit = str(self.config.get('record_limit', API_RECORD_LIMIT)).strip()
+        try:
+            # Defualt record limit will be set for None and Whitespaces
+            # Whitespaces before and after will be trimmed around valid numeric strings
+            return int(literal_eval(record_limit) if record_limit.strip() else API_RECORD_LIMIT)
+        except (NameError, SyntaxError, ValueError) as e:
+            raise ValueError("Invalid numeric value: " + str(self.config.get('record_limit'))) from e
+
+    def send_request_get_results(self, req, endpoint, params, count, **kwargs):
+        # Increament timeout duration based on retry attempt
+        resp = session.send(req, timeout=Stream.request_retry_count*self.request_timeout)
 
         # Sleep for provided time and retry if rate limit exceeded
         if 'Too Many Requests' in resp.reason:
@@ -149,6 +181,7 @@ class Stream():
             raise Server42xRateLimitError(resp.reason)
 
         resp.raise_for_status() # Check for requests status and raise exception in failure
+        Stream.request_retry_count = 1    # Reset retry count after success
 
         dec = humps.decamelize(resp.json())
         return dec
@@ -160,10 +193,13 @@ class Stream():
                           giveup=lambda e: e.response is not None and 400 <= e.
                           response.status_code < 500,
                           factor=BACKOFF_FACTOR,
-                          jitter=None)
+                          jitter=None,
+                          on_backoff=retry_handler,
+                          on_giveup=reset_request_retry_count)
     @backoff.on_exception(backoff.expo, (ConnectionError, ProtocolError, ReadTimeoutError), # backoff error
                           max_tries=7,
-                          factor=2)
+                          factor=2,
+                          jitter=None)
     @tap_pendo_utils.ratelimit(1, 2)
     def request(self, endpoint, params=None, count=1, **kwargs):
         # Set requests headers, url, methods, params and extra provided arguments
@@ -257,7 +293,7 @@ class Stream():
     def load_schema(self):
         refs = self.load_shared_schema_refs() # Load references scheamas
 
-        schema_file = "schemas/{}.json".format(self.name)
+        schema_file = f"schemas/{self.name}.json"
         with open(get_abs_path(schema_file)) as f:
             schema = json.load(f)
         self.resolve_schema_references(schema, "$ref", refs)
@@ -334,7 +370,6 @@ class Stream():
         # On ressuming, once replication of bookmarked parent_id is done, for next parent_id replication should start from original bookmark
         self.update_bookmark(state=state, stream=sub_stream.name, bookmark_value=previous_sync_completed_ts, bookmark_key="previous_sync_completed_ts")
 
-
     def sync_substream(self, state, parent, sub_stream, parent_response):
         # If last sync was interrupted, get last processed parent record
         last_processed = self.get_last_processed(state, sub_stream)
@@ -378,8 +413,16 @@ class Stream():
 
                 # Filtering the visitors based on last updated to improve performance of visitor_history replication
                 if isinstance(parent, Visitors):
+                    if record['metadata'].get('pendo', {'donotprocess': False}).get('donotprocess'):
+                        # If any visitor is set the Do Not Process flag then Pendo will stop collecting events
+                        # from that visitor and its records will be missing required keys, so we must skip it.
+                        LOGGER.info("Record marked as 'Do Not Process': %s", record[parent.key_properties[0]])
+                        continue
+
                     last_processed = self.get_last_processed(state, sub_stream)
-                    parent_last_updated = datetime.fromtimestamp(float(record['metadata']['auto'][self.replication_key]) / 1000.0, timezone.utc)
+                    # If there is no last_update key then set 'last_updated=now()' and fetch the records from recent bookmark
+                    parent_last_updated = datetime.fromtimestamp(float(record['metadata']['auto'].get(
+                        self.replication_key, now().timestamp() * 1000)) / 1000.0, timezone.utc)
 
                 if isinstance(parent, Visitors) and bookmark_dttm > parent_last_updated + timedelta(days=1):
                     LOGGER.info("No new updated records for visitor id: %s", record[parent.key_properties[0]])
@@ -403,12 +446,11 @@ class Stream():
                                 # Visitor_history API extracts records per day so we will extract all the records at once
                                 loop_for_records = False
                             else:
-                                stream_events, loop_for_records = sub_stream.sync(state, bookmark_dttm,
-                                                                                  record.get(parent.key_properties[0]))
+                                (_, stream_events), loop_for_records = sub_stream.sync(state, bookmark_dttm,
+                                                                                       record.get(parent.key_properties[0]))
 
                             # Loop over data of sub-stream
                             for event in stream_events:
-                                counter.increment()
 
                                 # Get metadata for the stream to use in transform
                                 schema_dict = sub_stream.stream.schema.to_dict()
@@ -434,8 +476,18 @@ class Stream():
                                 replication_value = transformed_record.get(sub_stream.replication_key)
                                 if replication_value:
                                     event_time = strptime_to_utc(replication_value)
+
+                                    # visitor_history stream replicates duplicate records older than bookmark value
+                                    # Skip older events than the previous sync
+                                    if isinstance(parent, Visitors):
+                                        previous_sync_completed_time = strptime_to_utc(
+                                            previous_sync_completed_ts) - timedelta(days=self.lookback_window())
+                                        if event_time < previous_sync_completed_time:
+                                            continue
+
                                     new_bookmark = max(new_bookmark, event_time)
 
+                                counter.increment()
                                 singer.write_record(sub_stream.stream.tap_stream_id,
                                                     transformed_record)
 
@@ -463,40 +515,166 @@ class Stream():
                                            previous_sync_completed_ts=strftime(final_bookmark))
         update_currently_syncing(state, None)
 
+    def get_pipeline_key_index(self, body, search_key):
+        for index, param in enumerate(body['request']['pipeline']):
+            if list(param.keys())[0] == search_key:
+                return index
+        return None
+
+    def get_body(self, key_id=None, period=None, first=None):
+        """This method will be overriden in the child class"""
+        return {}
+
+    def set_request_body_filters(self, body, start_time, records=None):
+        """Sets the filter parameter in the request body"""
+        # Note: even after we set filter value as first, while processing it will be set to the nearest day/hour range slot
+        # so we need to provide record filter to avoid any duplicate record
+        # Also we are using limit parameter as well which takes first N records for processing, rest records get discarded
+        # Considering this we may need to increase record limit in case record limit has reached in last response
+
+        limit_index = self.get_pipeline_key_index(body, 'limit')
+        filter_index = self.get_pipeline_key_index(body, 'filter')
+
+        body['request']['pipeline'][limit_index]['limit'] = self.record_limit
+        if isinstance(self, Accounts):
+            replication_key = 'metadata.auto.lastupdated'
+            replication_key_value = records[-1]['metadata']['auto']['lastupdated'] if records and len(records) > 0 else None
+        else:
+            replication_key = humps.camelize(self.replication_key)
+            replication_key_value = records[-1].get(humps.decamelize(self.replication_key)) if records and len(records) > 0 else None
+
+        # non-event based streams don't have filter parameter applied
+        if filter_index:
+            body['request']['pipeline'][filter_index]['filter'] = f'{replication_key}>={replication_key_value or start_time}'
+
+        return body
+
+    def remove_last_timestamp_records(self, records):
+        """Removes the overlapping records with last timestamp value. This avoids possibilty of duplicates"""
+        last_processed = []
+        decamalized_replication_key = humps.decamelize(self.replication_key)
+
+        if len(records) > 0:
+            if isinstance(self, Accounts):
+                timestamp = records[-1]['metadata']['auto']['lastupdated']
+                while records and timestamp == records[-1]['metadata']['auto']['lastupdated']:
+                    last_processed.append(records.pop())
+            else:
+                timestamp = records[-1].get(decamalized_replication_key)
+                while records and timestamp == records[-1].get(decamalized_replication_key):
+                    last_processed.append(records.pop())
+
+        # This is a corner cases where all records in the set have same timestamp
+        # This can occur if record limit is set very smaller compared to the max record limit
+        # In this case we will try to set record limit to max limit to make it harder to run into this issue
+        # But still can't completely dismiss the minor possibility of this issue occurring
+        if len(records) == 0 or not last_processed:
+            self.record_limit = API_RECORD_LIMIT
+
+        return last_processed
+
+    def get_first_parameter_value(self, body):
+        return body['request']['pipeline'][0]['source']['timeSeries'].get('first', 0)
+
+    def set_time_series_first(self, body, records, first=None):
+        """Sets the timeSeries 'first' parameter in request body"""
+        if len(records) > 1:
+            # This condition considers that within current time window there could be some more records
+            # So we will set last record timestamp as first of time series
+            # Note: even if we set browser time as first but while processing it will set to nearest day/hour range slot
+            body['request']['pipeline'][0]['source']['timeSeries']['first'] = records[-1].get(self.replication_key)
+        elif first:
+            body['request']['pipeline'][0]['source']['timeSeries']['first'] = self.get_first_parameter_value(body)
+        else:
+            body['request']['pipeline'][0]['source']['timeSeries']['first'] = int(datetime.now().timestamp() * 1000)
+
+        return body
 
     def sync(self, state, start_date=None, key_id=None, parent_last_updated=None):
-        stream_response = self.request(self.name, json=self.get_body())['results'] or []
+        update_currently_syncing(state, self.name)
 
-        # Get and intialize sub-stream for the current stream
         if STREAMS.get(SUB_STREAMS.get(self.name)):
             sub_stream = STREAMS.get(SUB_STREAMS.get(self.name))(self.config)
         else:
             sub_stream = None
 
+        # If last processed records exists, then set first to timestamp of first record
+        if self.last_processed:
+            first = self.last_processed[0][humps.decamelize(self.replication_key)]
+        else:
+            first = int(start_date.timestamp()) * 1000
+
+        # Setup body for first request
+        body = self.get_body()
+        self.set_request_body_filters(body, first, [])
+
+        stream_records = []
+        loop_for_records = False
+        while True:
+            # Loop breaks when paged response returns lesser records than set record limit
+            records = self.request(self.name, json=body).get('results') or []
+
+            # Set first and filters for next request
+            self.set_request_body_filters(
+                body,
+                first,
+                records)
+
+            if len(records) > 1:
+                # Previously removed records get duplicated in subsequent api response which needs to be removed
+                if self.last_processed:
+                    decamelized_replication_key = humps.decamelize(self.replication_key)
+                    records = [record for record in records if record.get(
+                        decamelized_replication_key) >= self.last_processed[0].get(decamelized_replication_key)]
+
+                removed_records = self.remove_last_timestamp_records(records)
+                stream_records += records
+
+                if self.last_processed == removed_records:
+                    stream_records += removed_records
+                    self.last_processed = None
+                    break
+
+                self.last_processed = removed_records
+
+            elif len(records) <= 1:
+                stream_records += records
+                self.last_processed = None
+                break
+
+            # If record limit set is reached, then return the extracted records
+            # Set the last processed records to ressume the extraction
+            if len(stream_records) >= self.record_limit:
+                loop_for_records = True
+                break
+
+        # These is a corner cases where this limit may get changed so reseeting it before next iteration
+        self.record_limit = self.get_default_record_limit()
+
         # Sync substream if the current stream has sub-stream and selected in the catalog
-        if stream_response and sub_stream and sub_stream.is_selected():
-            self.sync_substream(state, self, sub_stream, stream_response)
+        if stream_records and sub_stream and sub_stream.is_selected():
+            self.sync_substream(state, self, sub_stream, stream_records)
 
         update_currently_syncing(state, None)
-        return (self.stream, stream_response)
+        return (self.stream, stream_records), loop_for_records
 
     def lookback_window(self):
         # Get lookback window from config and verify value
         lookback_window = self.config.get('lookback_window') or '0'
         if not lookback_window.isdigit():
-            raise TypeError("lookback_window '{}' is not numeric. Check your configuration".format(lookback_window))
+            raise TypeError(f"lookback_window '{lookback_window}' is not numeric. Check your configuration")
         return int(lookback_window)
 
+
 class LazyAggregationStream(Stream):
     def send_request_get_results(self, req, endpoint, params, count, **kwargs):
         # Set request timeout to config param `request_timeout` value.
         # If value is 0,"0", "" or None then it will set default to default to 300.0 seconds if not passed in config.
-        config_request_timeout = self.config.get('request_timeout')
-        request_timeout = config_request_timeout and float(config_request_timeout) or REQUEST_TIMEOUT # pylint: disable=consider-using-ternary
         try:
             # Send request for stream data directly(without 'with' statement) so that
             # exception handling and yielding can be utilized properly below
-            resp = session.send(req, stream=True, timeout=request_timeout)
+            # Increament timeout duration based on retry attempt
+            resp = session.send(req, stream=True, timeout=count*self.request_timeout)
 
             if 'Too Many Requests' in resp.reason:
                 retry_after = 30
@@ -547,7 +725,7 @@ class LazyAggregationStream(Stream):
             # which flush out during sync_substream call above
             stream_response = self.request(self.name, json=self.get_body()) or []
 
-        return (self.stream, stream_response)
+        return (self.stream, stream_response), False
 
 class EventsBase(Stream):
     DATE_WINDOW_SIZE = 1
@@ -561,20 +739,11 @@ class EventsBase(Stream):
         self.replication_key = "day" if self.period == 'dayRange' else "hour"
         self.last_processed = None
 
-        # Record limit will throttle the number of records getting replicated
-        # This limit will resolve request timeouts and will reduce the peak memory consumption
-        record_limit = str(self.config.get('record_limit', API_RECORD_LIMIT)).strip()
-        try:
-            # Defualt record limit will be set for None and Whitespaces
-            # Whitespaces before and after will be trimmed around valid numeric strings
-            self.record_limit = int(literal_eval(record_limit) if record_limit.strip() else API_RECORD_LIMIT)
-        except (NameError, SyntaxError, ValueError) as e:
-            raise ValueError("Invalid numeric value: " + str(self.config.get('record_limit'))) from e
 
     def get_body(self, key_id, period, first):
         """This method returns generic request body of events steams"""
 
-        sort = humps.camelize(self.replication_key)
+        sort_key = humps.camelize(self.replication_key)
         return {
             "response": {
                 "mimeType": "application/json"
@@ -589,15 +758,12 @@ class EventsBase(Stream):
                                 "last": "now()"
                             }
                         }
-                    },
-                    {
-                        "sort": [sort]
-                    },
-                    {
+                    }, {
+                        "sort": [sort_key]
+                    }, {
+                        "filter": f"{sort_key}>=1"
+                    }, {
                         "limit": self.record_limit
-                    },
-                    {
-                        "filter": ""
                     }
                 ]
             }
@@ -606,53 +772,6 @@ class EventsBase(Stream):
     def get_first_parameter_value(self, body):
         return body['request']['pipeline'][0]['source']['timeSeries'].get('first', 0)
 
-    def set_time_series_first(self, body, records, first=None):
-        """Sets the timeSeries 'first' parameter in request body"""
-        if len(records) > 1:
-            # This condition considers that within current time window there could be some more records
-            # So we will set last record timestamp as first of time series
-            # Note: even if we set browser time as first but while processing it will set to nearest day/hour range slot
-            body['request']['pipeline'][0]['source']['timeSeries']['first'] = records[-1].get(self.replication_key)
-        elif first:
-            body['request']['pipeline'][0]['source']['timeSeries']['first'] = self.get_first_parameter_value(body)
-        else:
-            body['request']['pipeline'][0]['source']['timeSeries']['first'] = int(datetime.now().timestamp() * 1000)
-
-        return body
-
-    def set_request_body_filters(self, body, start_time, records=None):
-        """Sets the filter parameter in the request body"""
-        # Note: even if we set browser time as first but while processing it will set to nearest day/hour range slot
-        # so we need to provide record filter to avoid any duplicate replications
-        # Also we are using limit parameter as well which takes first N records for processing, rest records get discarded
-        # Considering this we may need to increase record limit in case record limit has reached in last response
-        camalized_replication_key = humps.camelize(self.replication_key)
-        body['request']['pipeline'][2]['limit'] = self.record_limit
-        if records and len(records) > 0:
-            # If there are 5 times records of record limits, in that case limit parameter will be increased acordiingly
-            body['request']['pipeline'][3]['filter'] = f'{camalized_replication_key}>={records[-1].get(self.replication_key)}'
-        else:
-            body['request']['pipeline'][3]['filter'] = f'{camalized_replication_key}>={start_time}'
-
-        return body
-
-    def remove_last_timestamp_records(self, records):
-        """Removes the overlapping records with last timestamp value. This avoids possibilty of duplicates"""
-        last_processed = []
-        if len(records) > 0:
-            timestamp = records[-1].get(self.replication_key)
-            while records and timestamp == records[-1].get(self.replication_key):
-                last_processed.append(records.pop())
-
-        # This is a corner cases where all records in the set have same timestamp
-        # This can occur if record limit is set very smaller compared to the max record limit
-        # In this case we will try to set record limit to max limit to make it harder to run into this issue
-        # But still can't completely dismiss the minor possibility of this issue occurring
-        if len(records) == 0 or not last_processed:
-            self.record_limit = API_RECORD_LIMIT
-
-        return last_processed
-
     def sync(self, state, start_date=None, key_id=None, parent_last_updated=None):
         update_currently_syncing(state, self.name)
 
@@ -665,11 +784,11 @@ class EventsBase(Stream):
 
         # If last processed records exists, then set first to timestamp of first record
         try:
-            first = self.last_processed[0][self.replication_key] if self.last_processed else int(lookback.timestamp()) * 1000
+            first = self.last_processed[0][humps.decamelize(
+                self.replication_key)] if self.last_processed else int(lookback.timestamp()) * 1000
         except Exception as e:
             LOGGER.info(str(e))
 
-
         # Setup body for first request
         body = self.get_body(key_id, period, first)
         self.set_time_series_first(body, [], first)
@@ -681,38 +800,52 @@ class EventsBase(Stream):
             records = self.request(self.name, json=body).get('results') or []
             self.set_time_series_first(body, records)
 
-            # Set first and filters for next request
-            self.set_request_body_filters(
-                body,
-                self.get_first_parameter_value(body),
-                records)
-
             if len(records) > 1:
-                removed_records = self.remove_last_timestamp_records(records)
-                events += records
-
-                if self.last_processed == removed_records:
-                    events += removed_records
+                if len(records) < self.record_limit:
+                    # If response returns less records than record limit means there are no more records to sync
+                    events += records
                     self.last_processed = None
                     break
 
-                self.last_processed = removed_records
+                removed_records = self.remove_last_timestamp_records(records)
+                if len(records) > 0:
+                    events += records
+
+                    if self.last_processed == removed_records:
+                        events += removed_records
+                        self.last_processed = None
+                        break
+
+                    self.last_processed = removed_records
+                else:
+                    # This block handles race condition where all records have same replication key value
+                    first = self.last_processed[0][humps.decamelize(
+                        self.replication_key)] if self.last_processed else int(lookback.timestamp()) * 1000
+
+                    body = self.get_body(key_id, period, first)
+                    continue
 
             elif len(records) == 1:
                 events += records
                 self.last_processed = None
                 break
 
+            # Set first and filters for next request
+            self.set_request_body_filters(
+                body,
+                self.get_first_parameter_value(body),
+                records)
+
             # If record limit set is reached, then return the extracted records
             # Set the last processed records to ressume the extraction
             if len(events) >= self.record_limit:
-                return events, True
+                return (self.stream, events), True
 
         # These is a corner cases where this limit may get changed so reseeting it before next iteration
-        self.record_limit = int(self.config.get('record_limit', API_RECORD_LIMIT))
+        self.record_limit = self.get_default_record_limit()
 
         update_currently_syncing(state, None)
-        return events, False
+        return (self.stream, events), False
 
 
 class Accounts(Stream):
@@ -721,7 +854,7 @@ class Accounts(Stream):
     replication_key = "lastupdated"
     key_properties = ["account_id"]
 
-    def get_body(self):
+    def get_body(self, key_id=None, period=None, first=None):
         return {
             "response": {
                 "mimeType": "application/json"
@@ -731,7 +864,13 @@ class Accounts(Stream):
                 "pipeline": [{
                     "source": {
                         "accounts": None
-                    }
+                    },
+                }, {
+                    "sort": ["metadata.auto.lastupdated"]
+                }, {
+                    "filter": "metadata.auto.lastupdated>=1"
+                }, {
+                    "limit": self.record_limit
                 }],
                 "requestId": "all-accounts",
                 "sort": ["accountId"]
@@ -756,7 +895,7 @@ class Features(Stream):
     replication_method = "INCREMENTAL"
     replication_key = "lastUpdatedAt"
 
-    def get_body(self):
+    def get_body(self, key_id=None, period=None, first=None):
         return {
             "response": {
                 "mimeType": "application/json"
@@ -769,7 +908,9 @@ class Features(Stream):
                         "features": None
                     }
                 }, {
-                    "sort": ["id"]
+                    "sort": [f"{self.replication_key}"]
+                }, {
+                    "limit": self.record_limit
                 }],
                 "requestId":
                 "all-features"
@@ -788,7 +929,7 @@ class FeatureEvents(EventsBase):
         return body
 
 
-class Events(LazyAggregationStream):
+class Events(EventsBase):
     name = "events"
     DATE_WINDOW_SIZE = 1
     key_properties = ['visitor_id', 'account_id', 'server', 'remote_ip']
@@ -800,23 +941,6 @@ class Events(LazyAggregationStream):
         self.period = config.get('period')
         self.replication_key = "day" if self.period == 'dayRange' else "hour"
 
-    def sync(self, state, start_date=None, key_id=None, parent_last_updated=None):
-        update_currently_syncing(state, self.name)
-
-        bookmark_date = self.get_bookmark(state, self.name,
-                                          self.config.get('start_date'),
-                                          self.replication_key)
-        bookmark_dttm = strptime_to_utc(bookmark_date)
-
-        # Set lookback window
-        lookback = bookmark_dttm - timedelta(
-            days=self.lookback_window())
-
-        # get events data
-        events = self.get_events(lookback, state, bookmark_dttm)
-        update_currently_syncing(state, None)
-        return (self.stream, events)
-
     def get_events(self, window_start_date, state, bookmark_dttm):
         # initialize start date as max bookmark
         max_bookmark = bookmark_dttm
@@ -879,28 +1003,14 @@ class Events(LazyAggregationStream):
     def transform(self, record):
         return humps.decamelize(record)
 
-    def get_body(self, period, first, end):
-        return {
-            "response": {
-                "mimeType": "application/json"
-            },
-            "request": {
-                "pipeline": [{
-                    "source": {
-                        "events": None,
-                        "timeSeries": {
-                            "period": period,
-                            "first": first,
-                            "last": end
-                        }
-                    }
-                }, {
-                    "sort": [self.replication_key]
-                }]
-            }
-        }
+    def get_body(self, key_id, period, first):
+        body = super().get_body(key_id, period, first)
+        body['request']['pipeline'][0]['source'].update({"events": None})
+        return body
 
-class PollEvents(Stream):
+
+
+class PollEvents(EventsBase):
     replication_method = "INCREMENTAL"
     name = "poll_events"
     key_properties = ['visitor_id', 'account_id', 'server_name', 'remote_ip']
@@ -911,47 +1021,11 @@ class PollEvents(Stream):
         self.period = config.get('period')
         self.replication_key = 'browser_time'
 
-    def get_body(self, period, first):
-        sort = humps.camelize(self.replication_key)
-        return {
-            "response": {
-                "mimeType": "application/json"
-            },
-            "request": {
-                "pipeline": [{
-                    "source": {
-                        "pollEvents": None,
-                        "timeSeries": {
-                            "period": period,
-                            "first": first,
-                            "last": "now()"
-                        }
-                    }
-                }, {
-                    "sort": [sort]
-                }]
-            }
-        }
+    def get_body(self, key_id, period, first):
+        body = super().get_body(key_id, period, first)
+        body['request']['pipeline'][0]['source'].update({"pollEvents": None})
+        return body
 
-    def sync(self, state, start_date=None, key_id=None, parent_last_updated=None):
-        update_currently_syncing(state, self.name)
-
-        # Get bookmark from state or start date for the stream
-        bookmark_date = self.get_bookmark(state, self.name,
-                                          self.config.get('start_date'),
-                                          self.replication_key)
-        bookmark_dttm = strptime_to_utc(bookmark_date)
-
-        # Set lookback window
-        lookback = bookmark_dttm - timedelta(
-            days=self.lookback_window())
-        ts = int(lookback.timestamp()) * 1000
-
-        # Get period type from config and make request for event's data
-        period = self.config.get('period')
-        body = self.get_body(period, ts)
-        events = self.request(self.name, json=body).get('results') or []
-        return self.stream, events
 
 class TrackEvents(EventsBase):
     replication_method = "INCREMENTAL"
@@ -984,7 +1058,7 @@ class TrackTypes(Stream):
     replication_method = "INCREMENTAL"
     replication_key = "lastUpdatedAt"
 
-    def get_body(self):
+    def get_body(self, key_id=None, period=None, first=None):
         return {
             "response": {
                 "mimeType": "application/json"
@@ -996,7 +1070,9 @@ class TrackTypes(Stream):
                         "trackTypes": None
                     }
                 }, {
-                    "sort": ["id"]
+                    "sort": [f"{self.replication_key}"]
+                }, {
+                    "limit": self.record_limit
                 }],
                 "requestId": "all-track-types"
             }
@@ -1008,7 +1084,7 @@ class Guides(Stream):
     replication_method = "INCREMENTAL"
     replication_key = "lastUpdatedAt"
 
-    def get_body(self):
+    def get_body(self, key_id=None, period=None, first=None):
         return {
             "response": {
                 "mimeType": "application/json"
@@ -1021,7 +1097,9 @@ class Guides(Stream):
                         "guides": None
                     }
                 }, {
-                    "sort": ["id"]
+                    "sort": [f"{self.replication_key}"]
+                }, {
+                    "limit": self.record_limit
                 }],
                 "requestId":
                 "all-guides"
@@ -1034,7 +1112,7 @@ class Pages(Stream):
     replication_method = "INCREMENTAL"
     replication_key = "lastUpdatedAt"
 
-    def get_body(self):
+    def get_body(self, key_id=None, period=None, first=None):
         return {
             "response": {
                 "mimeType": "application/json"
@@ -1047,7 +1125,9 @@ class Pages(Stream):
                         "pages": None
                     }
                 }, {
-                    "sort": ["id"]
+                    "sort": [f"{self.replication_key}"]
+                }, {
+                    "limit": self.record_limit
                 }],
                 "requestId":
                 "all-pages"
@@ -1146,7 +1226,7 @@ class Visitors(LazyAggregationStream):
     replication_key = "lastupdated"
     key_properties = ["visitor_id"]
 
-    def get_body(self):
+    def get_body(self, key_id=None, period=None, first=None):
         include_anonymous_visitors = self.config.get('include_anonymous_visitors') or DEFAULT_INCLUDE_ANONYMOUS_VISITORS
         anons = str(include_anonymous_visitors).lower() == 'true'
         return {
@@ -1190,7 +1270,7 @@ class MetadataAccounts(Stream):
     # the endpoint attribute overriden and re-initialized with different endpoint URL and method
     endpoint = Endpoints("/api/v1/metadata/schema/account", "GET")
 
-    def get_body(self):
+    def get_body(self, key_id=None, period=None, first=None):
         return None
 
     def sync(self, state, start_date=None, key_id=None, parent_last_updated=None):
@@ -1220,7 +1300,7 @@ class MetadataVisitors(Stream):
     # the endpoint attribute overriden and re-initialized with different endpoint URL and method
     endpoint = Endpoints("/api/v1/metadata/schema/visitor", "GET")
 
-    def get_body(self):
+    def get_body(self, key_id=None, period=None, first=None):
         return None
 
     def sync(self, state, start_date=None, key_id=None, parent_last_updated=None):
